#include "ground.h"
#define SDS
#include "alloc.h"
#include "simglob.h"
#include "norm.h"
#include "linklist.h"

linklist_t *site_storage=NULL;

#if PARALLEL==1
linklist_t ****linklist(vector *f1p,vector *f2p,vector *rp)
#else
linklist_t ****linklist(vector *fp,vector *rp)
#endif
/*
 * A link cell list {No.cell[0] * No.cell[1] * No.cell[2]} of the whole
   configuration is built according to the x,y,z coordinates of sites.
 * Memory needed is allocated in the 1st pass and re-used later.
 * x,y,z are normalized to [0,L),
 * type (linklist_t****) is interpreted as
   array[No.cell[0]][No.cell[1]][No.cell[2]] of type (linklist_t*)
 */
{
  int i,sp,ix,iy,iz,n,ns;
  static int oldncell[3],oldns=0;
  static linklist_t ****list=NULL;
  linklist_t *l;
  vector rL;
#if PARALLEL==1
  vector *r,*f1,*f2;
#else
  vector *r,*f;
#endif
#ifdef POLAR
  vector *rpol;
#if PARALLEL==1
  vector *f1pol,*f2pol;
#else
  vector *fpol;
#endif
#endif
  molecule_t *mn;

  ix=0;
  loop (i,0,DIM) {
    if (No.cell[i]<2) ERROR(("invalid No.cell[%d]=%d",i,No.cell[i]))
    if (No.cell[i]!=oldncell[i]) ix++; }

  VVV(rL,=No.cell,/box.L)

  if (!site_storage) {
    /* all memory for sites allocated in advance */
    allocarray(site_storage,No.s);
    oldns=No.s; }
  else
    if (oldns!=No.s)
      ERROR(("linklist: No.s=%d has changed (old=%d)",No.s,oldns))

  if (ix && list) {
    prt("number of cells changed => link-cell list reallocated");
    loop (ix,0,oldncell[0]) {
      loop (iy,0,oldncell[1]) free(list[ix][iy]);
      free(list[ix]); }
    free(list); /* note that my free clears the pointer to NULL ! */ }

  if (!list) {
    /* 3D array of cells allocated */
    alloc(list,No.cell[0]*sizeof(linklist_t***));
    loop (ix,0,No.cell[0]) {
      alloc(list[ix],No.cell[1]*sizeof(linklist_t**));
      loop (iy,0,No.cell[1])
	alloc(list[ix][iy],No.cell[2]*sizeof(linklist_t*)); }
    VV(oldncell,=No.cell) }

  /* clear pre-allocated 3D array of cells */
  loop (ix,0,No.cell[0])
    loop (iy,0,No.cell[1])
      loop (iz,0,No.cell[2]) list[ix][iy][iz]=NULL;

  l=site_storage;

  loop (n,0,No.N) if ((sp=(mn=molec+n)->sp)>=0) {
    ns=mn->ns;
    r=rof(mn,rp);
#if PARALLEL==1
    f1=rof(mn,f1p);
    f2=rof(mn,f2p);
#else
    f=rof(mn,fp);
#endif
#ifdef POLAR
    rpol=(vector*)((char*)r+polar_off);
#if PARALLEL==1
    f1pol=(vector*)((char*)f1+polar_off);
    f2pol=(vector*)((char*)f2+polar_off);
#else
    fpol=(vector*)((char*)f+polar_off);
#endif
#endif
    loop (i,0,ns) {
      /* normalize to [0,L) */
#ifdef WORM
#define IF while
#else
#define IF if
#endif
      l->r[0]=r[i][0];
      IF (l->r[0]>=box.L[0]) l->r[0]-=box.L[0];
      l->r[1]=r[i][1]; IF (l->r[1]>=box.L[1]) l->r[1]-=box.L[1];
      l->r[2]=r[i][2];
#ifndef WALL
                       IF (l->r[2]>=box.L[2]) l->r[2]-=box.L[2];
#endif
#define CELL(IX,I) \
      if (box.r0>0) \
        if (l->r[I]<-box.r0) { WARNING(("linkcell: %d.%d %c=%g < -box.r0\n*** box.r0 reset: old=%g, new=%g\n\
*** it is recommended to set box.rmin in the data\n\
*** (box.r0 is a growable copy of user-given box.rmin)",n,i,'x'+I,l->r[I],box.r0,-l->r[I]*1.1)) box.r0=-l->r[I]*1.1; } \
      IX=(int)(l->r[I]*rL[I]); \
      if (IX>=No.cell[I]) IX--;

/* removed: */
/*         if (l->r[I]>box.L[I]+box.r0) { WARNING(("linkcell: mol.site=%d.%d %c=%g > L+box.r0\n*** (box.r0=%g doubled)",n,i,'x'+I,l->r[I],box.r0)) box.r0*=2; } } \ */
      CELL(ix,0)
      CELL(iy,1)
      CELL(iz,2)
#ifdef POLAR
      VV(l->rpol,=rpol[i])
#if PARALLEL==1
      l->f1pol=f1pol[i];
      l->f2pol=f2pol[i];
#else
      l->fpol=fpol[i];
#endif
#endif
#if PARALLEL==1
      l->f1=f1[i];
      l->f2=f2[i];
#else
      l->f=f[i];
#endif
      l->sp=sp;
      l->n=n;
      l->si=&(spec[sp]->si[i]);
      l->next=list[ix][iy][iz];
      list[ix][iy][iz]=l;
      l++; } }

  if (box.r0>box.r0limit) {
#define R0LIMIT 4
    if (box.r0>R0LIMIT) ERROR(("limit box.r0=%g for sites in the basic cell is too large\n\
*** check box.rmin, tau.rho, tau.P, norm",box.r0))
    else WARNING(("limit box.r0=%g for sites in the basic cell is too large\n\
*** check box.rmin, tau.rho, tau.P, norm\n\
*** (warning limit increased)",box.r0))
           box.r0limit*=2;
    Min(box.r0limit,R0LIMIT) }

  if (No.occup) {
    static int *hist,nhist;
    int n;

    if (!nhist) {
      nhist=No.occup;
      if (nhist<2) nhist=No.s*3/PROD(No.cell)+3;
      allocarray(hist,nhist+1); }

    arrayzero(hist,nhist+1);

    loop (ix,0,No.cell[0])
      loop (iy,0,No.cell[1])
        loop (iz,0,No.cell[2]) {
          n=0;
	  for (l=list[ix][iy][iz]; l; l=l->next) n++;
	  Min(n,nhist)
	  hist[n]++; }
    prt("cell occupancy:");
    loopto (n,0,nhist) if (hist[n]) prt("%3d  %d",n,hist[n]); }

  return list;
} /* linklist */


#if 0
int testlist(char *msg,linklist_t ***list) /*********************** testlist */
/* # of sites in the list is returned */
{
  int i,j,s=0;
  linklist_t *l;
  loop (i,0,No.cell[0])
    loop (j,0,No.cell[1])
      for(l=list[i][j]; l; l=l->next) {
	s++;
	if (l->m->st<0 || l->m->st>NSITES) {
	  debugf("error %s %i %i %i %6.3f %6.3f %6.3f",
		 msg,l->m->st,l->m->indx,l->m->desttr,
		 l->m->r[0],l->m->r[1],l->m->r[2]);
	  Error(""); }
        }

  return s;
} /* testlist */
#endif

void lcsetup(vector L) /******************************************** lcsetup */
/* called from main */
{
  int i,prd=1,xyz=0;
  double x,V=PROD(L);

  if (No.percell<=0) No.percell=5;

  underline("linked-cell list setup");

  x=pow(No.s/(V*No.percell),1./3);
  loop (i,0,DIM) {
    if (No.cell[i]==0) {
      No.cell[i]=x*L[i]+0.6;
      xyz |= 1<<i;
      Max(No.cell[i],2) }
    prd*=No.cell[i]; }

  prt("No.percell=%g -> %g, used to set No.cell in coordinates: %s",
      No.percell,(double)No.s/prd,xyz?prtxyz(xyz):"none");
  prt("No.cell=(%d,%d,%d), (%.3g,%.3g,%.3g) in cutoff, (%.3g,%.3g,%.3g) final",
      No.cell[0],No.cell[1],No.cell[2],
      box.cutoff*No.cell[0]/box.L[0],box.cutoff*No.cell[1]/box.L[1],box.cutoff*No.cell[2]/box.L[2],
      box.cutoff*No.cell[0]/    L[0],box.cutoff*No.cell[1]/    L[1],box.cutoff*No.cell[2]/    L[2]);

  prt("parameter to predict 1-4 interaction radius over14=%g",over14);

  if (box.r0limit==0) box.r0limit=0.5;

  if (box.rmin<0) {
    box.r0=0;
    prt("box.rmin<0 specified => box.rmin:=0 (all coordinates normalized from zero)");
    if (tau.P!=0 || tau.rho!=0 || option('m')>2 || No.lone)
      WARNING(("You specified box.rmin<0 with variable box, Gear, or L-dependants.\n\
*** Some non-bonded interactions for larger distances will be omitted.")) }
  else if (box.rmin>0) {
    box.r0=box.rmin;
    prt("box.rmin=%g specified",box.rmin); }
  else {
    if (tau.P==0 && tau.rho==0 && option('m')==2) box.r0=0;
    if (tau.P!=0 || tau.rho!=0) box.r0=0.05;
    if (option('m') || No.lone) box.r0=0.15;
    prt("NOTE: box.rmin has not been specified, I guess box.rmin=%g",box.r0); }
  prt("Explanation: the linked-cell list method requires that all coordinates are\n\
positive, but some methods (variable box, Gear prediction) may slightly violate\n\
this condition. box.rmin (internally box.r0) is the allowed margin.");
  _n

}
